try:
    from ExtensionClass import ExtensionClass
except:
    ExtensionClass = None

from new import classobj
from kjbuckets import *

def isAspect(ob):
    return isinstance(ob,_Aspect)

from TW.Features import _DefaultFeatureMap, ClassFeature

DefaultVerticalContext = lambda x:x

def _makeClass(name,bases,dict):

    for b in bases:
        if type(b) is ExtensionClass:
            return ExtensionClass(name,bases,dict)

    return classobj(name,bases,dict)


def _elimDupes(seq):
    d = {}; o=[]; have=d.has_key; add=o.append
    for s in seq:
        if not have(s): add(s); d[s]=1
    return o












class ComponentMaker:

    flavor        = None
    module        = None
    featureMapper = _DefaultFeatureMap
    
    def __init__(self,**kw):
    
        self.output       = {}
        self.globals      = {}
        self.transforms   = []
        self.features     = {}
        self.featureList  = []
        self.superclasses = []
        self.dependencies = kjSet()
        
        self.__dict__.update(kw)
        self.__args__ = kw
        if self.flavor is not None: self.dependencies.add(self.flavor)

    def finishComponent(self):

        features = self.features
        
        for k in self.dependencyOrder():
            features[k].finishFeature(k,self)

        for t in self.transforms:
            t.finishTransform(self)
            
        return self.makeComponent()


    def __getitem__(self,name):       return self.output[name]
    def __setitem__(self,name,value): self.output[name]=value
    def __delitem__(self,name):       del self.output[name]
    def has_key(self,name):           return self.output.has_key(name)
    def get(self,name,default=None):  return self.output.get(name,default)



    def addSuper(self,klassOrName):
        self.superclasses.append(klassOrName)
        if klassOrName is str(klassOrName): self.dependencies.add(klassOrName)

    def insertSupers(self,classList):
        self.superclasses[0:0] = list(classList)
        for c in classList:
            if c is str(c): self.dependencies.add(c)

    def makeComponent(self):    
        bases = _elimDupes(self.superclasses) or filter(None,[self.flavor])
        bases = tuple(map(lambda x,d=self.globals: d.get(x,x), bases))
        # XXX need detection of stray stringType bases here
        
        name = self.name
        dict = self.output
        module = self.module
        
        if module:
            dict = dict.copy()
            dict['__module__']=module
            
        klass = _makeClass(name, bases, dict)

        if module:
            from sys import modules
            modules[module].__dict__[self.name] = klass

        return klass


    def childMaker(self,name,**kw):
        args = self.__args__.copy()
        args['name'] = "%s.%s" % (self.name,name)
        args['globals'] = self.output
        args.update(kw)

        return apply(self.__class__,(),args)



    def dependencyGraph(self):
        features = self.features
        g   = kjGraph()
        add = g.add
        
        for name,feature in features.items():
            for dep in feature.localDependencies(name,self): add(name,dep)
            
        return g            

    def dependencyOrder(self):
    
        features = self.features
        g = self.dependencyGraph()
        
        g1 = g.tclosure()
        for name in g.values():
            if g1.member(name,name):
                raise TypeError, ("Member %s is part of a dependency cycle for component %s" % (name,self.name))

        order  = []
        froms  = kjSet(g.keys())
        tos    = kjSet(g.values())
        others = kjSet(features.keys()) - (froms+tos)
        leaves = (tos - froms + others).items()
        
        while g:
            starts  = froms - tos
            
            for node in starts.items():
                order.append(node)
                del g[node]
                
            froms = kjSet(g.keys())
            tos   = kjSet(g.values())
                
        order.extend(leaves)
        order.reverse()
        return filter(features.has_key,order)


    def addFeature(self,name,item,vmap=DefaultVerticalContext):

        features = self.features

        if features.has_key(name):
            features[name].weaveFeature(name,item,self,vmap)
        else:
            self.featureMapper(item).weaveFeature(name,item,self,vmap)

        self.featureList.append(name)
        

    def addTransform(self, transform, dict=None, vmap=DefaultVerticalContext):
        self.transforms.append(transform)
        transform.begin(self,dict,vmap)


    def overlayDictionary(self,dict):
    
        verticals = {}; transform = None

        for k,v in dict.items():
            verticals[id(v)]=k

        vmap = lambda x,v=verticals: v.get(id(x),x)
        
        if dict.has_key('__features__'):
            transform = dict['__features__']
            dict=dict.copy()
            del dict['__features__']
            self.addTransform(transform,dict,vmap)
            
        for k,v in dict.items():
            self.addFeature(k,v,vmap)

        if transform is not None:
            transform.transform(self,dict,vmap)

    def isMaker(self,obj):
        return isinstance(obj,ComponentMaker)

class _Aspect:

    def __init__(self,name,bases=(),dict=None,__doc__=None):

        self.__bases__ = bases
        self.__name__  = name
        self.__doc__   = __doc__

        cbases = []
        abases = []

        _Aspect = self.__class__
        dicts = filter(None,[dict])

        for b in bases:
            if isinstance(b,_Aspect):
                dicts.extend(b.dictionaries)
                cbases.extend(list(b.componentBases))
            else:
                cbases.append(b)

        self.componentBases = tuple(cbases)
        self.dictionaries   = dicts

        if __doc__ is None:
            for d in dicts:
                if d.has_key('__doc__'):
                    self.__doc__ = d['__doc__']; break

    def __add__(self,other):
        return _Aspect(self.__name__,(other,self))  # XXX need to handle other being class better
        
    def __getitem__(self,key):
        for d in self.dictionaries:
            if d.has_key(key): return d[key]
        raise KeyError,key

    def has_key(self,key):
        for d in self.dictionaries:
            if d.has_key(key): return 1

    def keys(self):
        k = kjSet()
        for d in self.dictionaries:
            k=k+kjSet(d.keys())
        return k.items()
        
    def weave(self, makerType=ComponentMaker, *args, **kw):
    
        kwd = {'name':self.__name__}
        kwd.update(kw)
        
        maker = apply(makerType,args,kwd)
        self.weaveInto(maker)
        return maker.finishComponent()

        
    def weaveInto(self,maker,verticalContext=DefaultVerticalContext):
        maker.insertSupers(map(verticalContext,self.componentBases))
        dicts = self.dictionaries[:]
        dicts.reverse()
        map(maker.overlayDictionary,dicts)


    __call__ = weave
    
    def __repr__(self):
        return "<Aspect %s at %x>" % (self.__name__, id(self))


Aspect = _Aspect("Aspect")











class AspectMaker(ComponentMaker):

    def makeComponent(self):
        bases = tuple(_elimDupes(self.superclasses) or filter(None,[self.flavor]))
        return _Aspect(self.name, bases, self.output)

    def addClass(self,name,flavor=None,supers=(),featureItems=()):
    
        c = self[name] = self.childMaker(name, flavor=flavor)
        
        if supers:
            c.insertSupers(supers)
        
        for n,f in featureItems:
            c.addFeature(n,f)
            
        self.features[name]=ClassFeature
        return c

    def addFeature(self,name,item,vmap=DefaultVerticalContext):
        self[name] = item
        


def parentComponentClassOf(obj):
    import string,sys
    name   = string.join(string.split(obj.__name__,'.')[:-1],'.')
    module = sys.modules[obj.__module__]
    return getattr(module, name, None)
