from xml.sax import ContentHandler, parse

class ObjectMakingHandler(ContentHandler):

    """SAX handler that makes a pseudo-DOM"""
    
    def __init__(self,documentClass):
        self.stack = []
        self.documentClass=documentClass
        ContentHandler.__init__(self)
        
    def startDocument(self):
        self.stack.append(self.documentClass())

    def startElement(self, name, atts):
        top = self.stack[-1]
        node = top._newNode(name,atts)
        self.stack.append(node)

    def characters(self, ch):
        self.stack[-1]._addText(ch)

    def endElement(self, name):    
        stack = self.stack
        top = stack.pop()
        
        if top._name != name:
            raise SyntaxError,"End tag '%s' found when '%s' was wanted" % (name, top._name)
            
        out = top._finish()
        
        if out is not None:
            stack[-1]._addNode(name,out)

    def endDocument(self):
        document = self.document = self.stack[0]._finish()
        del self.stack




class Node:

    def __init__(self,name='',atts={}):
        self._name = name
        self._subNodes = []
        self._allNodes = []
        
        d=self.__dict__
        for a in atts.keys():
            d[a]=atts[a]
            
    def _addNode(self,name,node):
        self._allNodes.append(node)
        self._subNodes.append(node)
        d=self.__dict__
        if not d.has_key(name): d[name]=[]
        d[name].append(node)

    def _newNode(self,name,atts):
        return self.__class__(name,atts)
        
    def _addText(self,text):
        self._allNodes.append(text)

    def _get(self,name):
        return self.__dict__.get(name,[])

    def _findFirst(self,name):
        d=self._get(name)
        if d: return d
        for n in self._subNodes:
            if hasattr(n,'_findFirst'):
                d = n._findFirst(name)
                if d: return d
                
    def _finish(self): return self
    




class Document(Node):

    def _finish(self):
        self.documentElement = self._subNodes[0]
        return self
        
    def _newNode(self,name,atts):
        return Node(name,atts)



def load(filename_or_stream, documentClass=Document):
    handler = ObjectMakingHandler(documentClass)
    parse(filename_or_stream, handler)
    return handler.document

