import Interface

class IServiceInfo(Interface.Base):
    """Need to be able to find an element/feature by name"""

class IElementInfo(Interface.Base):
    """Need to be able to list features"""

class IFeatureInfo(Interface.Base):
    """Need to know multiplicity, composition, navigability, type, inverse relationship, etc."""





























    

class StructuralFeature(Interface.Base):

    def __call__(self):
        """Return the value of the feature"""
        return aq_acquire(self,'_getValueFor')(self)

    def values(self):
        """Return the value(s) of the feature as a sequence, even if it's a single value"""
        return aq_acquire(self,'_getValuesFor')(self)

    def set(self,value):
        """Set the value of the feature to value"""
        return aq_acquire(self,'_setValueFor')(self,value)

    def clear(self):
        """Unset the value of the feature (like __delattr__)"""
        return aq_acquire(self,'_clearValueFor')(self)

    def addItem(self,item):
        """Add the item to the collection/relationship, reject if multiplicity exceeded"""
        return aq_acquire(self,'_addItemFor')(self,item)

    def removeItem(self,item):
        """Remove the item from the collection/relationship, if present"""
        return aq_acquire(self,'_removeItemFor')(self,item)

    def addNewItem(self,*args,**kw):
        """Create a new object in the collection of the appropriate type, passing through args/kw
           and returning the newly created object."""
        return aq_acquire(self,'_addNewItemFor')(self,args,kw)

    def isEmpty(self):
        """Return true if the value is currently not set or the collection is empty"""
        return aq_acquire(self,'_valueIsEmptyFor')(self,args,kw)
        
    def marshal(self,*args,**kw):
        """Convert args[0] or kwargs into an 'item' suitable for use with the other methods.
           Raises ValueError if invalid format."""



    def _notifyLink(self,otherEnd):
        """Called to signal that this feature has been added to the feature 'otherEnd'"""
        
    def _notifyUnlink(self,otherEnd):
        """Called to signal that this feature has been removed from the feature 'otherEnd'"""




class AssociationEnd(StructuralFeature):
    pass

class FeatureImplementation(Interface.Base):

    def _addItemFor(self,feature,item):
        pass

    def _setValueFor(self,feature,item):
        pass
        
    def _clearValueFor(self,feature,item):
        pass

    def _removeItemFor(self,feature,item):
        pass

    def _addNewItemFor(self,feature,args,kw):
        pass

    def _getValueFor(self,feature):
        pass

    def _getValuesFor(self,feature):
        pass
