from kjbuckets import *
import TW.SOX
from TW.Aspects import AspectMaker
from TW.Features import FeatureExpression
from stat import ST_MTIME
from os import path, stat, unlink


def _upToDate(source,dest):
    return  source and dest and \
            path.exists(dest) and path.exists(source) and \
            stat(source)[ST_MTIME]<=stat(dest)[ST_MTIME] 


class AttributeFeature(FeatureExpression):

    def __init__(self,name,type,valueExpr="",argsExpr="",kind="attribute_feature"):
        self.dependencies = type,
        self.kind = kind
        expr = self.expr = "attribute_feature(%s,%s(%s),%s)" % (`name`,type,valueExpr,argsExpr)
        self._compile(expr)




















class MetaModelNode(TW.SOX.Node):

    contentMap = kjGraph([
        ('', 'metamodel'), ('metamodel','package'),
        ('package','package'),
        ('package','association'), ('association','role'),
        ('package','primitive'),
        ('package','enumeration'), ('enumeration','literal'),
        ('package','datatype'),
            ('datatype','attribute'),('datatype','superclass'),
        ('package','element'),
            ('element','attribute'),('element','superclass'),
    ])

    requiredAttribs = kjGraph([
        ('package','name'), ('primitive','name'), ('enumeration','name'),
        ('literal','name'), ('datatype','name'), ('element','name'),
        ('attribute','name'), ('attribute','type'), ('element','abstract'),
        ('role','name'), ('role','type'), ('role','composite'), ('role','kind'),
        ('role','navigable'), ('superclass','type'),
    ])

    allowedAttribs = kjGraph(requiredAttribs)
    allowedAttribs.add('role','xmi.name')
    name = ''

    def __init__(self,name='',atts={},factory=None,path=None):
    
        anames     = kjSet(atts.keys())
        disallowed = anames - self.allowedAttribs.reachable(name)
        missing    = self.requiredAttribs.reachable(name) - anames
        
        if missing:
            raise SyntaxError,"Missing attributes %s in element %s" % (missing.items(),name)
            
        if disallowed:
            raise SyntaxError,"Invalid attributes %s in element %s" % (disallowed.items(),name)
            
        TW.SOX.Node.__init__.im_func(self,name,atts or {})
        self.factory, self.path = factory, path

    def _newNode(self,name,atts):
        if not self.contentMap.member(self._name,name):
            reachable = self.contentMap.reachable(self._name).items()
            raise SyntaxError,"Found %s where %s expected" % (name, `reachable`)
        path = self.path
        if self._name not in ('','metamodel'):
            if path is None:
                path=''
            else:
                path='%s%s.' % (path,self.name)
        return MetaModelNode(name,atts,self.factory,path)

    def _finish(self): return getattr(self,'_finish_'+self._name)()

    def _finish_(self): return None

    def _finish_package(self):
        a = self._get('association')
        for p in self._get('package'): a.extend(p)
        return a

    def _setXMINames(self, maker):
        maker['_XMINames'] = (self.path+self.name),

    def _finish_primitive(self):
        self._setXMINames( self.factory.addClass(self.name,'primitive_flavor') )
        
    def _finish_enumeration(self):
        self._setXMINames( self.factory.addClass(self.name, 'enumeration_flavor', (), self._get('literal')) )

    def _finish_literal(self):
        return self.name, self.name

    def _finish_attribute(self):
        return self.name, AttributeFeature(self.name, self.type, '', `self.path+self.name`)

    def _finish_datatype(self):
        maker = self.factory.addClass(self.name, 'datatype_flavor', self._get('superclass'), self._get('attribute'))
        self._setXMINames(maker)


    def _finish_element(self):
        maker = self.factory.addClass(self.name, 'element_flavor', self._get('superclass'), self._get('attribute'))
        self._setXMINames(maker)
        maker['Abstract'] = self.abstract=='true'
        return None

    def _finish_superclass(self):
        return self.type

    def _finish_role(self):
        self.isParent=(self.composite=='true')
        self.kind = self.kind + '_feature'
        self.navigable=(self.navigable=='true')
        self.xmi_name=getattr(self,'xmi.name',self.name)
        self.proper_name=(self.kind=='ref') and self.xmi_name or self.factory.plurals.get(self.xmi_name,self.xmi_name)
        return self

    def _finish_association(self):
        roles = tuple(self._get('role'))
        if len(roles)<>2: raise SyntaxError,"Expected 2 roles, found %d" % len(roles)
        
        r0, r1 = roles; r0.home = r1.type; r1.home = r0.type
        r0.isChild=r1.isParent; r0.other_name=r1.proper_name; r0.other_key=r1.name; r0.reversible=r1.navigable
        r1.isChild=r0.isParent; r1.other_name=r0.proper_name; r1.other_key=r0.name; r1.reversible=r0.navigable

        return roles

    def _finish_metamodel(self):
        for alist in self._get('package'):
            for a in alist:
                for r in a:
                    if r.isParent:  r.kind='parent_'+r.kind
                    #elif r.isChild: r.kind='child_' +r.kind
                    r.fullPath = '%s.%s' % (self.factory[r.home]['_XMINames'][0], r.name)

                    argsExpr = '%s(%s,%s,%s,%s,%s)' % (r.kind,
                        `r.proper_name`, `r.type`, `r.fullPath`,
                        `r.other_name`, `r.navigable`
                    )
                    self.factory[r.home].addFeature(r.proper_name,FeatureExpression(argsExpr))
                    
class StructuralAspectMaker(AspectMaker):

    def addFeature(self,name,item,vmap=None):
        self[name] = item
        if not self.has_key('__featureNames__'):
            self['__featureNames__'] = ()
        self['__featureNames__'] = self['__featureNames__'] + (name,)


































class _Loader:

    NodeClass = MetaModelNode
        
    def __call__(self, modelFile, pluralsFile=None, cacheFile=None, makerType=StructuralAspectMaker, *args, **kw):

        if _upToDate(cacheFile,modelFile) and _upToDate(pluralsFile,cacheFile) and _upToDate(__file__,cacheFile):
            return self._loadFromCache(cacheFile)

        plurals = {}
        
        if pluralsFile: plurals = self._loadPlurals(pluralsFile)

        maker = apply(makerType,args,kw)
        maker.plurals = plurals
        
        TW.SOX.load(modelFile, lambda nc=self.NodeClass,m=maker: nc(factory=m))

        aspect = maker.finishComponent()

        if cacheFile is not None:
            self._saveToCache(cacheFile,aspect)

        return aspect
        
    def _loadPlurals(self,pluralsFile):
        plurals = {}
        from string import strip,split
        for l in open(pluralsFile).readlines():
            if '=' in l:
                [k,v]=split(strip(l),'=',1)
                plurals[k]=v
        return plurals
        
    def _loadFromCache(self,cacheFile):
        from cPickle import load
        return load(open(cacheFile,'r'))




    def _saveToCache(self, cacheFile, aspect):
        try:
            f = open(cacheFile,'w')
            try:
                from cPickle import dump
                dump(aspect,f,1)
            finally:
                f.close()
        except:
            try:
                unlink(cacheFile)
            except:
                pass
            raise

                
load = _Loader()



















