from unittest import TestCase, makeSuite, TestSuite
from types import ClassType
from TW.Aspects import *
from TW.Features import *
from Acquisition import Implicit
from ExtensionClass import ExtensionClass, Base
from kjbuckets import *


class ShuttleTests(TestCase):
    def setUp(self):
        self.s = Shuttle()
        
    

class aOne(Aspect):

    anAttr = 'aOne'

    class NestedBase(Implicit):
        nbAttr = 'nbOne'

class aTwo(Aspect):

    anAttr = 'aTwo'
    otherAttr = 'exists'

    class NestedBase(Base):
        nbAttr = 'nbTwo'

    class NestedDerived(NestedBase):
        ndAttr = 'ndTwo'

class aThree(Aspect,Implicit):

    anAttr = 'aThree'

    class NestedDerived:
        ndAttr = 'ndThree'


class Basics12(TestCase):

    component_name = 'c12'

    first_aspect = aOne
    second_aspect = aTwo

    def setUp(self):
        self.aspect    = self.first_aspect + self.second_aspect
        self.component = self.aspect(name=self.component_name, module=__name__)

    def checkName(self):
        c = self.component
        assert c.__name__ == self.component_name

    def checkAttr(self):
        a = self.second_aspect['anAttr']
        assert self.component.anAttr is a

    def checkInstalled(self):
        c = self.component_name
        assert globals().has_key(c)
        assert globals()[c] is self.component

    def checkExists(self):
        assert self.component.otherAttr == 'exists'

    def checkBases(self):
        cbases = self.component.__bases__
        abases = self.first_aspect.componentBases + self.second_aspect.componentBases
        for b in abases:
            assert b in cbases
        for b in cbases:
            assert b in abases

    def checkVerticalInheritance(self):
        assert self.component.NestedBase in self.component.NestedDerived.__bases__




    def checkNestedNames(self):
        c = self.component
        cn = self.component_name + '.'
        for n in 'NestedBase','NestedDerived':
            assert getattr(c,n).__name__ == cn+n

    def checkNestedModules(self):
        c = self.component
        for n in 'NestedBase','NestedDerived':
            assert getattr(c,n).__module__ == __name__

    def checkNestedInstalled(self):
        c = self.component
        cn = self.component_name + '.'
        for n in 'NestedBase','NestedDerived':
            assert globals().has_key(cn+n) and globals()[cn+n] is getattr(c,n)


    def checkNestedBases(self):
        c = self.component
        for n in 'NestedBase', 'NestedDerived':
            cbases = getattr(c,n).__bases__
            abases = []
            for a in self.first_aspect, self.second_aspect:
                for d in a.dictionaries:
                    if d.has_key(n): abases.extend(list(d[n].__bases__))

            aset = kjSet(abases)
            cset = kjSet(cbases)

            extras = cset - aset
            missing = aset - cset

            outputs = kjSet(c.__dict__.values())

            assert len(extras-outputs)==0
            assert len(missing)==len(extras)




    def checkNestedBaseAttr(self):
        if self.second_aspect.has_key('NestedBase'):
            b = self.second_aspect['NestedBase']
        else:
            b = self.first_aspect['NestedBase']
        assert self.component.NestedBase.nbAttr is b.nbAttr

    def checkNestedDerivedAttr(self):
        if self.second_aspect.has_key('NestedDerived'):
            b = self.second_aspect['NestedDerived']
        else:
            b = self.first_aspect['NestedDerived']
        assert self.component.NestedDerived.ndAttr is b.ndAttr


    def tearDown(self):
        del self.aspect; del self.component
        c = self.component_name
        if globals().has_key(c): del globals()[c]






















class Basics21(Basics12):
    component_name = 'c21'
    first_aspect = aTwo
    second_aspect = aOne

class Basics23(Basics12):
    component_name = 'c23'
    first_aspect = aTwo
    second_aspect = aThree

class Basics32(Basics12):
    component_name = 'c32'
    first_aspect = aThree
    second_aspect = aTwo

class Basics123(Basics12):
    component_name = 'c123'
    first_aspect = aOne
    second_aspect = aTwo + aThree

class Basics321(Basics12):
    component_name = 'c321'
    first_aspect = aThree
    second_aspect = aTwo + aOne

class Basics1_23(Basics12):
    first_aspect = aOne
    class second_aspect(aTwo,aThree):
        pass

class Basics3_21(Basics12):
    first_aspect = aThree
    class second_aspect(aTwo,aOne):
        pass







class Verticals(Aspect):

    class Root:
        pass

    class L1a(Root):
        pass

    class L1b(Root):
        pass

    class L2a(L1a):
        pass

    class L2b(L1b):
        pass

    class L3(L2a,L2b):
        pass


class Horizontals(Aspect):

    class L2a(Implicit):
        pass
















class TestVerticals(TestCase):

    def checkVerticals(self):
        c = Verticals()

        for ac in Verticals.__dict__.values():
            if type(ac) is ClassType:
                n = ac.__name__
                cc = getattr(c,n)
                cbases = map(lambda x: x.__name__,              cc.__bases__)
                abases = map(lambda x: 'Verticals.'+x.__name__, ac.__bases__)
                assert cbases == abases

    #def checkHorizontals(self):
    #    try:
    #        (Verticals+Horizontals)()
    #    except TypeError,v:
    #        assert str(v)=='Mixed vertical and horizontal inheritance for component Verticals.L2a'
    #    else:
    #        assert None,"Did not get a type error for mixed inheritance"





















class FeatureTestMixin(Aspect):

    __features__ = FeatureSet(owClass = OverwriteFeature)

    class owClass:
        attr = "owClass"


class FeatureTestBase(Aspect):

    class owClass(Implicit):
        attr = "this should not be here"


class FeatureTestMixed1(FeatureTestMixin, FeatureTestBase): pass
class FeatureTestMixed2(FeatureTestBase, FeatureTestMixin): pass


class FeatureTest(TestCase):

    def checkOverwrite(self):
        mix = FeatureTestMixed1()
        assert mix.owClass is FeatureTestMixin['owClass']

    def checkUnderwrite(self):
        mix = FeatureTestMixed2()
        assert mix.owClass is FeatureTestBase['owClass']














class A1(Aspect):

    attr1 = 'A1-attr1'

    class B1:
        baz='spam'
        spanish='inquisition'

    class B2(Implicit):
        pass

    class B3(B2,B1):

        class B3Sub:
            thing = 'B3Sub1'

class A2(Aspect):

    attr1 = 'A2-attr1'

    class B1:
        foo='bar'
        spanish='unexpected'

    class B3:
        spammish='inquisition'

        class B3Sub:
            thing = 'B3Sub2'

class A3(Aspect,Implicit):
    pass


class C(A1,A2): pass
class D(A2,A1): pass
class E(A1,A2,A3): pass
class F(A3,A1,A2): pass



class OldTests(TestCase):

    def setUp(self):
        self.C = C() # W.weave('C',C)
        self.D = D() # W.weave('D',D)
        self.E = E() # W.weave('E',E)
        self.F = F() # W.weave('F',F)

    def checkA3(self):
        assert A3.componentBases==(Implicit,)
        assert type(A3.componentBases[0]) is ExtensionClass

    def checkB3(self):
        for c in self.C, self.D:
            assert type(c.B3) is ExtensionClass
            assert c.B3.__bases__ == (c.B2,c.B1)
            assert c.B3.spammish =='inquisition'

    def checkNestedClasses(self):
        assert self.C.B3.B3Sub.thing=='B3Sub1'
        assert self.D.B3.B3Sub.thing=='B3Sub2'

    def checkNewClasses(self):
        assert self.C.B1 is not A1.dictionaries[0]['B1']
        assert self.C.B2 is not A1.dictionaries[0]['B2']
        assert self.C.B2.__bases__[0]==Implicit
        assert self.C.B3 is not A1.dictionaries[0]['B3']

    def checkEF(self):
        assert E.componentBases==(Implicit,)
        assert F.componentBases==(Implicit,)

    def checkC(self):
        assert self.C.__name__ == 'C'
        assert self.C.attr1 == 'A1-attr1'
        assert type(self.C) is ClassType
        assert self.C.__bases__==()




    def checkD(self):
        assert self.D.attr1 == 'A2-attr1'
        assert self.D.__bases__==()

    def checkE(self):
        assert self.E.__name__ == 'E'
        assert self.E.attr1 == 'A1-attr1'
        assert type(self.E) is ExtensionClass
        assert (self.E.__bases__ == (Implicit,)), str(self.E.__bases__)
    
    def checkF(self):
        assert self.F.__name__ == 'F'
        assert self.F.attr1 == 'A1-attr1'
        assert type(self.F) is ExtensionClass
        assert (self.F.__bases__ == (Implicit,)), str(self.F.__bases__)

    def checkB1(self):
        for c in self.C, self.D:
            assert type(c.B1) is ClassType
            assert c.B1.foo=='bar'
            assert c.B1.baz=='spam'
        assert self.C.B1.spanish=='inquisition'
        assert self.D.B1.spanish=='unexpected'


















TestClasses = (
    ShuttleTests, Basics12, Basics21, Basics23, Basics32, Basics123, Basics321, Basics1_23, Basics3_21,
    TestVerticals, FeatureTest, OldTests
)

def suite():
    s = []
    for t in TestClasses:
        s.append(makeSuite(t,'check'))

    return TestSuite(s)

